"""
The qiprofile imaging Mongodb data model.
"""

import re
import decimal
from numbers import Number
import mongoengine
from mongoengine import (fields, signals, ValidationError)
from .common import (Encounter, Outcome, TumorExtent)


class Point(mongoengine.EmbeddedDocument):
    """The 3D point in the volume voxel space."""
    
    x = fields.IntField()
    """
    The x dimension value in the image coordinate system.
    """
    
    y = fields.IntField()
    """
    The y dimension value in the image coordinate system.
    """
    
    z = fields.IntField()
    """
    The z dimension value in the image coordinate system.
    """


class Image(mongoengine.EmbeddedDocument):
    """The image file encapsulation."""
    
    meta = dict(allow_inheritance=True)
    
    name = fields.StringField(required=True)
    """
    The image file base name. The client has the responsibility
    of determining the image file based on the image store. For example,
    a NIfTI scan volume XNAT 1.6 archive 3D volume image location is as
    # follows::
          
          /path/to/archive/<project>/arc001/<experiment>/SCANS/<scan>/NIFTI/<name>
      
      where:
      
      * *project* is the XNAT project name
      * *experiment* is the XNAT experiment label
      * *scan* is the scan number
      * *name* is the volume image file base name
    """
    
    metadata = fields.DictField()
    """Additional image properties, e.g. average intensity."""


class LabelMap(Image):
    """A label map with an optional associated color lookup table."""
    
    color_table = fields.StringField()
    """
    The color map lookup table file base name relative to the XNAT
    archive ROI resource location.
    """


class Region(mongoengine.EmbeddedDocument):
    """The 3D region in volume voxel space."""
    
    mask = fields.EmbeddedDocumentField(Image)
    """
    The binary mask file relative to the image store ROI resource
    location.
    """
    
    resource = fields.StringField(required=True)
    """The region imaging store resource name, e.g. ``roi``."""
    
    label_map = fields.EmbeddedDocumentField(LabelMap)
    """The region overlay :class:`LabelMap` object."""
    
    centroid = mongoengine.EmbeddedDocumentField(Point)
    """The region centroid."""


class Resource(mongoengine.EmbeddedDocument):
    """The image store file access abstraction."""
    
    meta = dict(allow_inheritance=True)
    
    name = fields.StringField(required=True)
    """The image store name used to access the resource."""


class SingleImageResource(Resource):
    """A resource with one file."""
    
    meta = dict(allow_inheritance=True)
    
    image = fields.EmbeddedDocumentField(Image)
    """The sole resource image."""


class MultiImageResource(Resource):
    """A resource with several files."""
    
    meta = dict(allow_inheritance=True)
    
    images = fields.ListField(
        field=fields.EmbeddedDocumentField(Image)
    )
    """The resource images."""


class ImageSequence(Outcome):
    """The Scan or Registration."""
    
    meta = dict(allow_inheritance=True)
    
    time_series = fields.EmbeddedDocumentField(SingleImageResource)
    """The 4D time series resource."""
    
    volumes = fields.EmbeddedDocumentField(MultiImageResource)
    """The 3D volumes resource."""


class Protocol(mongoengine.Document):
    """
    The image processing protocol abstract class.
    """
    
    meta = dict(allow_inheritance=True, collection='qiprofile_protocol')
    
    technique = fields.StringField(required=True)
    """
    The image acquisition or processing technique, e.g. ``T1`` for a
    T1-weighted scan or 'ANTs' for an ANTs registration.
    
    The REST update client is responsible for ensuring that technique
    synonyms resolve to the same technique value, e.g. scans with
    descriptions including ``T1`` and ``T1 AXIAL`` should both resolve
    to technique ``T1``. The acquisition and processing details are
    stored in the configuration rather than embedded in the technique.
    """
    
    configuration = fields.DictField()
    """
    The image processing input parameter
    {*section*\ : {*option*\ : *value*\ }} dictionary,
    e.g.::
        
        {
          'fsl.FLIRT': {'bins': 640, 'cost_func': 'normcorr'},
          'fsl.FNIRT' : {'in_fwhm': [10,6,2,2], 'ref_fwhm': [10,6,2,2]}
        }
    
    The configuration should be sufficient to reproduce the image
    processing result in the context of the image processing pipeling.
    The sections are pipeline task interfaces or other pipeline
    input groupings.
    """


class RegistrationProtocol(Protocol):
    """The registration settings."""
    
    technique = fields.StringField()
    """
    The image processing technique, e.g. ``ANTs`` for registration. The
    REST update client is responsible for ensuring that technique synonyms
    resolve to the same technique value, e.g. scans with descriptions
    including ``BLISS`` and ``BLISS_AUTO_SENSE`` should both resolve to
    technique ``BLISS``.
    """
    
    configuration = fields.DictField()
    """
    The image processing input parameter
    {*section*\ : {*option*\ : *value*\ }} dictionary,
    e.g.::
        
        {
          'FLIRT': {'bins': 640, 'cost_func': 'normcorr'},
          'FNIRT' : {'in_fwhm': [10,6,2,2], 'ref_fwhm': [10,6,2,2]}
        }
    
    The configuration should be sufficient to reproduce the image
    processing result in the context of the image processing pipeling.
    The sections are pipeline task interfaces or other pipeline
    input groupings.
    
    :Note: MongoDB does not permit dotted dictionary keys. Thus, e.g.,
         'fsl.FNIRT' is not allowed in the preceding example.
    """


class Registration(ImageSequence):
    """
    The patient image registration that results from processing a scan.
    """
    
    protocol = fields.ReferenceField(RegistrationProtocol, required=True)
    """The registration protocol."""


class ScanProtocol(Protocol):
    """
    The scan acquisition protocol. Scans with the same protocol
    and image dimensions are directly comparable, e.g. in comparing
    modeling results across subjects or sessions.
    
    The recommended technique controlled values include, but are not
    limited to, the following:
    * ``T1`` - T1-weighted
    * ``T2`` - T2-weighted
    * ``DW`` - diffusion-weighted
    * ``PD`` - proton density
    """
    pass


class Scan(ImageSequence):
    """
    The the concrete subclass of the abstract :class:`ImageSequence`
    class for scans.
    """
    
    number = fields.IntField(required=True)
    """
    The scan number. In the XNAT image store, each scan is
    identified by a number unique within the session.
    """
    
    protocol = fields.ReferenceField(ScanProtocol, required=True)
    """The scan acquisition protocol."""
    
    preview = fields.StringField()
    """
    The image file base name relative to the image store scan
    preview resource location.
    """
    
    bolus_arrival_index = fields.IntField()
    """
    The bolus arrival volume index, or None if this is not a
    DCE scan.
    """
    
    rois = fields.ListField(field=fields.EmbeddedDocumentField(Region))
    """
    The image regions of interest. For a scan with ROIs, there is
    one ROI per scan tumor. The rois list order is the same as the
    :class:`qirest-client.model.clinical.PathologyReport`
    ``tumors`` list order.
    """
    
    registrations = fields.ListField(
        field=fields.EmbeddedDocumentField(Registration)
    )
    """
    The registrations performed on the scan.
    """


class ModelingProtocol(Protocol):
    """The modeling input options."""
    pass


class Modeling(Outcome):
    """
    The QIN pharmicokinetic modeling run on an image sequence.
    """
    
    class ParameterResult(mongoengine.EmbeddedDocument):
        """The output for a given modeling run result parameter."""
        
        image = fields.EmbeddedDocumentField(Image)
        """
        The voxel-wise mapping file name relative to the XNAT
        modeling archive resource location. The image metadata
        should include *average*, the average modeling result
        value across all voxels.
        """
        
        label_map = fields.EmbeddedDocumentField(LabelMap)
        """The label map overlay NIfTI file."""
    
    class Source(mongoengine.EmbeddedDocument):
        """
        This Modeling.Source embedded class works around the following
        mongoengine limitation:
        
        * mongoengine does not allow heterogeneous collections, i.e.
          a domain model Document subclass cannot have subclasses.
          Furthermore, the domain model Document class cannot be
          an inner class, e.g. ModelingProtocol.
        
        Consequently, the Modeling.source field cannot represent an
        abstract superclass of ScanProtocol and RegistrationProtocol.
        This Source embedded document introduces a disambiguation
        level by creating a disjunction object that can either hold
        a *scan reference or a *registration* reference.
        """
        
        scan = fields.ReferenceField(ScanProtocol)
        
        registration = fields.ReferenceField(RegistrationProtocol)
    
    protocol = fields.ReferenceField(ModelingProtocol, required=True)
    """The modeling protocol."""
    
    source = fields.EmbeddedDocumentField(Source, required=True)
    """
    The modeling source protocol.
    
    Since a given :class`Session` contains only one :class:`ImageSequence`
    per source protocol, the image sequence on which modeling is performed
    is determined by the source protocol. Specifying the source as a
    protocol rather than the specific scan or registration allows modeling
    to be embedded in the :class`Session` document rather than the
    :class:`SessionDetail`.
    """
    
    resource = fields.StringField(required=True)
    """The modeling imaging store resource name, e.g. ``pk_R3y9``."""
    
    result = fields.DictField(
        field=mongoengine.EmbeddedDocumentField(ParameterResult)
    )
    """
    The modeling {*parameter*: *result*} dictionary, where:
    
    - *parameter* is the lower-case underscore parameter key, e.g.
      ``k_trans``.
    
    - *result* is the corresponding :class:`ParameterResult`
    
    The parameters are determined by the :class:`ModelingProtocol`
    technique. For example, the `OHSU QIN modeling workflow`_ includes
    the following outputs for the FXL (`Tofts standard`_) model and the
    FXR (`shutter speed`_) model:
    
    - *fxl_k_trans*, *fxr_k_trans*: the |Ktrans| vascular permeability
       transfer constant
    
    - *delta_k_trans*: the FXR-FXL |Ktrans| difference
    
    - *fxl_v_e*, *fxr_v_e*: the |ve| extravascular extracellular volume
       fraction
    
    - *fxr_tau_i*: the |taui| intracellular |H2O| mean lifetime
    
    - *fxl_chi_sq*, *fxr_chi_sq*: the |chisq| intensity goodness of fit
    
    The REST client is responsible for anticipating and interpreting the
    meaning of the *parameter* based on the modeling technique. For
    example, if the image store has a session modeling resource
    ``pk_h7Jtl`` which includes the following files::
        
        k_trans.nii.gz
        k_trans_overlay.nii.gz
        chi_sq.nii.gz
        chi_sq_overlay.nii.gz
    
    then a REST database update client might calculate the average |Ktrans|
    and |chisq| values and populate the REST database as follows::
        
        from qirest_client.helpers import database
        t1 = database.get_or_create(ScanProtocol, dict(scan_type='T1'))
        tofts = database.get_or_create(ModelingProtocol,
                                       dict(technique='Tofts'))
        ktrans_label_map = LabelMap(filename='k_trans_overlay.nii.gz',
                                    color_table='jet.txt')
        ktrans = Modeling.ParameterResult(name='k_trans.nii.gz',
                                          average=k_trans_avg,
                                          label_map=ktrans_label_map)
        chisq_label_map = LabelMap(filename='k_trans_overlay.nii.gz',
                                   color_table='jet.txt')
        chisq = Modeling.ParameterResult(name='chi_sq.nii.gz',
                                         average=chi_sq_avg,
                                         label_map=chisq_label_map)
        result = dict(ktrans=ktrans, chisq=chisq)
        session.modeling = Modeling(protocol=tofts, source=t1,
                                    resource='pk_h7Jtl', result=result)
    
    It is then the responsibility of an imaging web app REST read client
    to interpret the modeling result dictionary items and display them
    appropriately.
    
    .. reST substitutions:
    .. include:: <isogrk3.txt>
    .. |H2O| replace:: H\ :sub:`2`\ O
    .. |Ktrans| replace:: K\ :sup:`trans`
    .. |ve| replace:: v\ :sub:`e`
    .. |taui| replace:: |tau|\ :sub:`i`
    .. |chisq| replace:: |chi|\ :sup:`2`
    
    .. _OHSU QIN modeling workflow: http://qipipe.readthedocs.org/en/latest/api/pipeline.html#modeling
    .. _Tofts standard: http://onlinelibrary.wiley.com/doi/10.1002/(SICI)1522-2586(199909)10:3%3C223::AID-JMRI2%3E3.0.CO;2-S/abstract
    .. _shutter speed: http://www.ncbi.nlm.nih.gov/pmc/articles/PMC2582583
    """
    
    def __str__(self):
        return "Modeling %s" % self.resource


class SessionDetail(mongoengine.Document):
    """The MR session detailed content."""
    
    meta = dict(collection='qiprofile_session_detail')
    
    scans = fields.ListField(field=mongoengine.EmbeddedDocumentField(Scan))
    """The list of scans."""


class Session(Encounter):
    """The MR session (a.k.a. *study* in DICOM terminology)."""
    
    modelings = fields.ListField(
        field=fields.EmbeddedDocumentField(Modeling)
    )
    """The modeling performed on the session."""
    
    tumor_extents = fields.ListField(
        field=fields.EmbeddedDocumentField(TumorExtent)
    )
    """The tumor extents measured from the scan."""
    
    detail = fields.ReferenceField(SessionDetail)
    """The session detail reference."""
    
    @classmethod
    def pre_delete(cls, sender, document, **kwargs):
        """Cascade delete this Session's detail."""
        
        if self.detail:
            self.detail.delete()

signals.pre_delete.connect(Session.pre_delete, sender=Session)



"""
The qiprofile unit of measurement MongoDB data model.

This uom module is an advisory module for quantity display and
conversion. Quantities are always stored in a canonical unit
documented in the field.
"""

import decimal
from numbers import Number
import mongoengine
from mongoengine import fields


class Decimal(mongoengine.EmbeddedDocument):
    """
    This Decimal document class is a work-around for the broken
    MongoEngine DecimalField
    (cf. https://github.com/MongoEngine/mongoengine/issues?q=is%3Aissue+is%3Aopen+decimal).

    Decimal has a float value and an optional precision. The default
    precision is determined as follows:

    * If the Decimal value is initialized with a python ``decimal.Decimal``
      or string, then the default precision is the number of decimal places
      expressed in that value, e.g.::

          Decimal('1.24').precision #=> 2
          Decimal(decimal.Decimal('1.4')).precision #=> 1

    * If the Decimal value is initialized with an integer, then the default
            precision is zero, e.g.::

                Decimal(1).precision #=> 0
    """
    # TODO - migrate to DecimalField when it is fixed in MongoEngine.

    def __init__(self, *args, **kwargs):
        # The positional arguments will be replaced by keyword arguments.
        if args:
            value = args[0]
            if len(args) > 1:
                precision = args[1]
            else:
                precision = None
        else:
            value = kwargs.get('value')
            precision = kwargs.get('precision')

        # Convert the value to a float, if necessary.
        if value != None:
            value = float(value)
        # Set the value keyword argument.
        kwargs['value'] = value

        # If the precision is not specified, then determine a default if
        # possible.
        if precision == None:
            if isinstance(value, decimal.Decimal):
                value_s = str(value)
                if '.' in value_s:
                    _, precision_s = value_s.split('.')
                    precision = int(precision_s)
            elif isinstance(value, int):
                precision = 0

        # Set the precision keyword argument.
        kwargs['precision'] = precision

        super(Decimal, self).__init__(**kwargs)

    def canonical(self):
        """
        :return: the python decimal.Decimal equivalent of this Decimal
        """
        return decimal.Decimal(str(value))

    def __repr__(self):
        """
        :return: the value as a string with decimal places truncated
            by the precision
        """
        if self.precision == None:
            return str(self.value)
        elif self.precision == 0:
            return str(int(self.value))
        else:
            pat = '%.' + ("%s" % self.precision) + 'f'
            return pat % self.value

    value = fields.FloatField(required=True)

    precision = fields.IntField()


class Unit(mongoengine.EmbeddedDocument):
    """
    Unit is an abstract class for the supported measurement units.
    Each unit has a scaling factor with the default scaling factor
    defined in the concrete Unit subclass. Each Unit subclass has
    the following class variables:

    * ``BASE`` - the standard unscaled metric unit abbreviation,
      e.g. ``m`` for the meter Extent unit.

    * ``SCALES`` - the recommended scaling factors, e.g. ``m`` for
       milli and ``c`` for centi.

    The scales are a list in preference order, e.g. the ``Extent.SCALES``
    value ``['c', 'm']`` implies that the preferred extent unit display
    is ``cm`` (centimeter) and an edit form should show the scales ``c``
    and ``m``, in that order, defaulting to ``c``. A scale of None
    signifies the scaling factor 1, e.g. the  the ``Radiation.SCALES``
    value ``[None, 'c']`` implies that the preferred radiation unit
    display is ``Gy`` (Gray) and an edit form should show the scales
    blank and ``c``, in that order, defaulting to blank (scaling factor
    1).

    These class variables are advisory. The client is responsible for
    displaying the base and scales and converting from a scaled value
    to an unscaled value before saving the measurement to the database.
    """

    meta = dict(allow_inheritance=True)


class Extent(Unit):

    SCALES = ['c', 'm']

    BASE = 'm'
    """The meter designator."""

    scale = fields.StringField(default='c')
    """The default extent is centimeter."""


class Weight(Unit):

    SCALES = ['m', 'k']

    BASE = 'g'
    """The gram designator."""

    scale = fields.StringField(default='m')
    """The default weight is milligram."""


class Volume(Unit):

    SCALES = ['m']

    BASE = 'l'
    """The liter designator."""

    scale = fields.StringField(default='m')
    """The default volume is milliliter."""


class Radiation(Unit):
    """The radiation unit is always an unscaled Gray."""

    BASE = 'Gy'
    """The Gray designator."""

class Measurement(mongoengine.EmbeddedDocument):
    """
    A scientific measurement.

    The measurement is a quantitative amount associated with a unit.
    The unit is the Unit as captured and displayed. The amount is
    expressed as a python Decimal in unscaled units. If the constructor
    is called with a non-Decimal numeric amount, then the value is
    converted to a Decimal, e.g.::

        Measurement(amount=0.006, unit=Weight())

    is equivalent to::

        from decimal import Decimal
        Measurement(amount=Decimal(0.006), unit=Weight())

    The measurement unit can be qualified by a second ``per_unit``
    dimension, e.g. 2 mg/kg dosage per patient weight is expressed
    as::

        Measurement(amount=0.002, unit=Weight(), per_unit=Weight(scale='k'))

    :Note: the amount is a :class:`Decimal` embedded object rather than
      the broken MongoEngine ``DecimalField`` (see the :class:`Decimal`
      comment).
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes the Measurement document. The amount can be either
        a positional or a keyword argument. The amount is converted to
        a Decimal.
        
        :param args: the amount as a positional argument
        :param kwargs: the following keyword arguments:
        :keyword amount: the amount as a keyword argument
        """
        # Convert the amount to a Decimal, if necessary. The amount
        # can be the first positional argument or a keyword argument.
        if args:
            args = list(args)
            args[0] = self._as_decimal(args[0])
        elif 'amount' in kwargs:
            kwargs['amount'] = self._as_decimal(kwargs['amount'])
        
        super(Measurement, self).__init__(*args, **kwargs)

    def _as_decimal(self, value):
        if isinstance(value, Number):
            return Decimal(value)
        else:
            return value

    amount = fields.EmbeddedDocumentField(Decimal, required=True)

    unit = fields.EmbeddedDocumentField(Unit)

    per_unit = fields.EmbeddedDocumentField(Unit)



from datetime import datetime
from mongoengine import (connect, ValidationError)
from nose.tools import (assert_true, assert_false, assert_equal,
                        assert_almost_equal, assert_raises)
from qirest_client.helpers import database
from qirest_client.model.subject import (Project, ImagingCollection, Subject)
from qirest_client.model.common import TumorExtent
from qirest_client.model.imaging import (
    Session, Scan, ScanProtocol, Registration, RegistrationProtocol,
    MultiImageResource, SingleImageResource, LabelMap, SessionDetail,
    Image, Point, Region, Modeling, ModelingProtocol
)
from qirest_client.model.clinical import (
    Biopsy, Evaluation, Surgery, PathologyReport, TumorLocation,
    TNM, BreastSurgery, BreastPathology, ResidualCancerBurden,
    ModifiedBloomRichardsonGrade, HormoneReceptorStatus,
    SarcomaPathology, FNCLCCGrade, NecrosisPercentValue,
    NecrosisPercentRange, necrosis_percent_as_score
)


class TestModel(object):
    """
    Basic data model test. A more complete test is found in the qirest
    server TestSeed test suite.
    """
    def setup(self):
        self._connection = connect(db='qiprofile_test')
        self._connection.drop_database('qiprofile_test')

    def tearDown(self):
        self._connection.drop_database('qiprofile_test')

    def test_project(self):
        prj = Project()
        # The project must have a name.
        with assert_raises(ValidationError):
            prj.validate()
        prj.name = 'Breast'
        prj.validate()

    def test_collection(self):
        # The collection must have a name.
        with assert_raises(ValidationError):
            ImagingCollection(project='QIN_Test').validate()
        # The collection must have a project.
        with assert_raises(ValidationError):
            ImagingCollection(name='Breast').validate()
        # A valid collection.
        coll = ImagingCollection(project='QIN_Test', name='Breast')
        coll.validate()

    def test_subject(self):
        subject = Subject()
        # The collection must have a project.
        with assert_raises(ValidationError):
            subject.validate()
        subject.project = 'QIN_Test'
        # The collection must have a number.
        with assert_raises(ValidationError):
            subject.validate()
        subject.number = 1
        # The subject must have a collection.
        with assert_raises(ValidationError):
            subject.validate()
        subject.collection = 'Breast'
        subject.validate()
    
    def test_race(self):
        subject = Subject(project='QIN_Test', collection='Breast', number=1)
        subject.races = ['White', 'Black', 'Asian', 'AIAN', 'NHOPI']
        subject.validate()
    
        subject = Subject(project='QIN_Test', collection='Breast', number=1)
        subject.races = ['Invalid']
        with assert_raises(ValidationError):
            subject.validate()
    
        # Races must be a list.
        subject.races = 'White'
        with assert_raises(ValidationError):
            subject.validate()
    
    def test_ethnicity(self):
        subject = Subject(project='QIN_Test', collection='Breast', number=1)
        subject.ethnicity = 'Non-Hispanic'
        subject.validate()
    
        # The ethnicity is a controlled value.
        subject.ethnicity = 'Invalid'
        with assert_raises(ValidationError):
            subject.validate()
    
    def test_breast_biopsy(self):
        """
        This Breast biopsy test case is a variation of the Breast
        surgery test case. Notably, this test case exercises multiple
        tumors. There is no Sarcoma biopsy test case.
        """
        subject = Subject(project='QIN_Test', collection='Breast', number=1)
        # The pathology.
        size = TNM.Size.parse('T3a')
        size.validate()
        grade = ModifiedBloomRichardsonGrade(
            tubular_formation=2, nuclear_pleomorphism=1, mitotic_count=2
        )
        grade.validate()
        tnm = TNM(tumor_type='Breast', grade=grade, size=size,
                  metastasis=False, resection_boundaries=1,
                  lymphatic_vessel_invasion=False)
        tnm.validate()
        estrogen1 = HormoneReceptorStatus(hormone='estrogen', positive=True,
                                          intensity=80)
        estrogen1.validate()
        hormone_receptors1 = [estrogen1]
        extent1 = TumorExtent(length=48, width=31, depth=19)
        extent1.validate()
        tumor1_pathology = BreastPathology(tnm=tnm, extent=extent1,
                                           hormone_receptors=hormone_receptors1)
        tumor1_pathology.validate()
        estrogen2 = HormoneReceptorStatus(hormone='estrogen', positive=False)
        estrogen2.validate()
        hormone_receptors2 = [estrogen2]
        extent2 = TumorExtent(length=27, width=16, depth=8)
        extent2.validate()
        tumor2_pathology = BreastPathology(extent=extent2,
                                           hormone_receptors=hormone_receptors2)
        tumor2_pathology.validate()
        # The pathology aggregate.
        pathology = PathologyReport(tumors=[tumor1_pathology, tumor2_pathology])
        pathology.validate()
        # Add the encounter to the subject.
        date = datetime(2013, 1, 4)
        biopsy = Biopsy(date=date, weight=54, pathology=pathology)
        biopsy.validate()
        subject.encounters = [biopsy]
        subject.validate()

    def test_rcb_index(self):
        extent = TumorExtent(length=32, width=12)
        rcb = ResidualCancerBurden(
                tumor_cell_density=40,
                dcis_cell_density=10,
                positive_node_count=6,
                largest_nodal_metastasis_length=8
            )
        path = BreastPathology(extent=extent, rcb=rcb)
        rcb_index = path.rcb_index()
        assert_almost_equal(rcb_index, 3.695, 3,
                            msg="The RCB index is incorrect: %f" % rcb_index)

    def test_rcb_class(self):
        path = BreastPathology()
        for expected, rcb_index in enumerate((0, 1.2, 1.4, 3.5)):
            actual = path.rcb_class(rcb_index)
            assert_equal(actual, expected,
                         "The RCB class of RCB index %f is incorrect: %d" % 
                        (rcb_index, actual))

    def test_breast_surgery(self):
        subject = Subject(project='QIN_Test', collection='Breast', number=1)

        # The pathology report.
        size = TNM.Size.parse('T2')
        size.validate()
        grade = ModifiedBloomRichardsonGrade(
            tubular_formation=1, nuclear_pleomorphism=1, mitotic_count=1
        )
        grade.validate()
        tnm = TNM(tumor_type='Breast', grade=grade, size=size,
                  metastasis=False, resection_boundaries=1,
                  lymphatic_vessel_invasion=False)
        tnm.validate()
        location = TumorLocation(sagittal_location='Left')
        location.validate()
        tumor_pathology = BreastPathology(tnm=tnm, location=location)
        tumor_pathology.validate()
        pathology = PathologyReport(tumors=[tumor_pathology])
        pathology.validate()

        # Add the encounter to the subject.
        date = datetime(2013, 1, 4)
        surgery = BreastSurgery(date=date, weight=54, surgery_type='Lumpectomy',
                               pathology=pathology)
        surgery.validate()
        subject.encounters = [surgery]
        subject.validate()
    
    def test_sarcoma_surgery(self):
        subject = Subject(project='QIN_Test', collection='Sarcoma', number=1)
        
        # The pathology report.
        size = TNM.Size.parse('T3a')
        size.validate()
        grade = FNCLCCGrade(
            differentiation=2, necrosis_score=1, mitotic_count=1
        )
        grade.validate()
        tnm = TNM(tumor_type='Sarcoma', grade=grade, size=size,
                  metastasis=False, resection_boundaries=1,
                  lymphatic_vessel_invasion=False)
        tnm.validate()
        location = TumorLocation(body_part='Thigh', sagittal_location='Left',
                                 coronal_location='Posterior')
        location.validate()
        tumor_pathology = SarcomaPathology(tnm=tnm, location=location)
        tumor_pathology.validate()
        pathology = PathologyReport(tumors=[tumor_pathology])
        pathology.validate()

        # Add the encounter to the subject.
        date = datetime(2014, 6, 19)
        surgery = Surgery(date=date, weight=47, pathology=pathology)
        surgery.validate()
        subject.encounters = [surgery]
        subject.validate()
    
    def test_tnm_size(self):
        for value in ['T1', 'Tx', 'cT4', 'T1b', 'cT2a']:
            size = TNM.Size.parse(value)
            size.validate()
            assert_equal(str(size), value, "The TNM parse is incorrect -"
                                           " expected %s, found %s"
                                           % (value, str(size)))

    def test_necrosis_score(self):
        fixture = {
            0: dict(integer=0,
                    value=NecrosisPercentValue(value=0),
                    range=NecrosisPercentRange(
                        start=NecrosisPercentRange.LowerBound(value=0),
                        stop=NecrosisPercentRange.UpperBound(value=1))),
            1: dict(integer=40,
                    value=NecrosisPercentValue(value=40),
                    range=NecrosisPercentRange(
                        start=NecrosisPercentRange.LowerBound(value=40),
                        stop=NecrosisPercentRange.UpperBound(value=50))),
            2: dict(integer=50,
                    value=NecrosisPercentValue(value=50),
                    range=NecrosisPercentRange(
                        start=NecrosisPercentRange.LowerBound(value=50),
                        stop=NecrosisPercentRange.UpperBound(value=60)))
        }
        for expected, inputs in fixture.iteritems():
            for in_type, in_val in inputs.iteritems():
                actual = necrosis_percent_as_score(in_val)
                assert_equal(actual, expected,
                             "The necrosis score for %s is incorrect: %d" %
                             (in_val, expected))
    
    def test_treatment(self):
       # TODO - add the treatment test case.
       pass
    
    def test_session(self):
        # The test subject.
        subject = Subject(project='QIN_Test', collection='Breast', number=1)
        # The test session.
        date = datetime(2013, 1, 4)
        session = Session(date=date)
        session.validate()
        subject.encounters = [session]
        subject.validate()
    
    def test_add_encounter(self):
        # The test subject.
        subject = Subject(project='QIN_Test', collection='Breast', number=1)
        # Add the test encounters.
        encounters = [Biopsy(date=datetime(2014, m, 1)) for m in (3, 5, 7)]
        subject.add_encounter(encounters[1])
        assert_equal(subject.encounters, encounters[1:2])
        subject.add_encounter(encounters[0])
        assert_equal(subject.encounters, encounters[0:2])
        subject.add_encounter(encounters[2])
        assert_equal(subject.encounters, encounters)
    
    def test_scan(self):
        # The scan protocol.
        protocol = database.get_or_create(ScanProtocol, dict(technique='T1'))
        # The scan.
        scan = Scan(protocol=protocol, number=1)
        scan.validate()
        # Validate the session detail embedded scan.
        detail = SessionDetail(scans=[scan])
        detail.validate()
    
    def test_registration(self):
        # The scan protocol.
        scan_pcl = database.get_or_create(ScanProtocol, dict(technique='T1'))
        # The scan.
        scan = Scan(protocol=scan_pcl, number=1)
        scan.validate()
        # The registration protocol.
        reg_pcl = database.get_or_create(RegistrationProtocol,
                                         dict(technique='FLIRT'))
        # The registration volumes.
        vol_imgs = [Image(name="volume00%s.nii.gz" % vol) for vol in range(1, 4)]
        volumes = MultiImageResource(name='reg', images=vol_imgs)
        # The registration time series.
        time_series_img = Image(name='reg_ts.nii.gz')
        time_series = SingleImageResource(name='reg_ts', image=time_series_img)
        reg = Registration(protocol=reg_pcl, volumes=volumes, time_series=time_series)
        reg.validate()
    
        # Validate the session detail and embedded scan registration.
        scan.registrations = [reg]
        detail = SessionDetail(scans=[scan])
        detail.scans = [scan]
        detail.validate()
    
    def test_roi(self):
        # The scan protocol.
        scan_pcl = database.get_or_create(ScanProtocol,
                                          dict(technique='T1'))
        # The scan.
        scan = Scan(protocol=scan_pcl, number=1)
        scan.validate()
    
        # The ROI.
        centroid = Point(x=200, y=230, z=400)
        intensity = 31
        metadata = dict(centroid=centroid, average_intensity=intensity)
        mask = Image(name='lesion1.nii.gz', metadata=metadata)
        label_map = LabelMap(name='lesion1_color.nii.gz',
                             color_table='color_table.nii.gz')
        label_map.validate()
        roi = Region(mask=mask, resource='roi', label_map=label_map, centroid=centroid)
        roi.validate()
        
        # Validate the session detail and embedded scan ROI.
        scan.rois = [roi]
        detail = SessionDetail(scans=[scan])
        detail.scans = [scan]
        detail.validate()
    
    def test_modeling_protocol(self):
        # The modeling protocol content.
        cfg = {'Fastfit': {'model_name': 'fxr.model'},
               'R1': {'r1_0_val': 0.7, 'baseline_end_idx': 1}}
        mdl_pcl_key = dict(technique='BOLERO', configuration=cfg)
        mdl_pcl = database.get_or_create(ModelingProtocol, mdl_pcl_key)
        assert_equal(mdl_pcl.configuration, cfg,
                     "The fetched modeling configuration is incorrect: %s" %
                     mdl_pcl.configuration)

    def test_modeling(self):
        # The test subject.
        subject = Subject(project='QIN_Test', collection='Breast', number=1)
        # The modeling protocol.
        cfg = {'Fastfit': {'model_name': 'fxr.model'},
               'R1': {'r1_0_val': 0.7, 'baseline_end_idx': 1}}
        mdl_pcl_key = dict(technique='BOLERO', configuration=cfg)
        mdl_pcl = database.get_or_create(ModelingProtocol, mdl_pcl_key)
        # The source protocol.
        scan_pcl = database.get_or_create(ScanProtocol, dict(technique='T1'))
        source = Modeling.Source(scan=scan_pcl)
        # The modeling data.
        ktrans_img = Image(name='ktrans.nii.gz')
        ktrans = Modeling.ParameterResult(image=ktrans_img)
        modeling = Modeling(protocol=mdl_pcl, source=source, resource='pk_01',
                            result=dict(ktrans=ktrans))
        modeling.validate()
        # Validate the subject embedded session modeling.
        date = datetime(2014, 3, 1)
        session = Session(date=date, modelings=[modeling])
        session.validate()
        subject.encounters = [session]
        subject.validate()

if __name__ == "__main__":
    import nose
    nose.main(defaultTest=__name__)



"""Mongo Engine interaction utilities."""

def get_or_create(klass, key, **non_key):
    """
    This function stands in for the Mongo Engine ``get_or_create``
    collection method which was deprecated in mongoengine v0.8.0
    and dropped in mongoengine v0.10.0, since MongoDB does not
    support transactions.
    
    If there is an object of the given Mongo Engine data model
    class which matches the given secondary key, then that object is
    returned. Otherwise, a new object is created with the content
    prescribed by both the primary and non-primary parameters.
    
    The create step is an upsert, i.e. a new object is created if
    and only if it does not yet exist. The upsert allows for the small
    possibility that an object is created after the fetch attempt
    but before the create attempt. In that situation, the existing
    object non-key content is modified and the modified object is
    returned.
    
    :Note: The idiom used in this function modifies the solution
        proposed in
        http://stackoverflow.com/questions/25846462/mongoengine-replacing-get-or-create-with-upsert-update-one/25863633#25863633.
        That StackOverflow work-around returns the following error:
    
            ValueError: update only works with $ operators

        The work-around to the StackOverflow work-around is to call
        the data model class *update_one* method rather than *modify*.

    :param klass: the Mongo Engine data model class
    :param key: the secondary key {attribute: value} dictionary
    :param non_key: the non-key {attribute: value} dictionary
    :return: the existing or new object
    """
    try:
        # Search by primary key.
        return klass.objects.get(**key)
    except klass.DoesNotExist:
        # Create the new object as an upsert. Specify the MongoDB Engine
        # set__*attribute* modification options for both the key and
        # non-key {attribute: value} dictionaries.
        content = key.copy()
        content.update(non_key)
        mod_opts = {'set__' + attr: val for attr, val in content.iteritems()}
        result = klass.objects(**key).update_one(upsert=True, full_result=True,
                                                 **mod_opts)
        # The object now exists, so fetch it by id.
        # The new object id is returned in the result upserted item.
        return klass.objects.get(id=result['upserted'])



"""
The qiprofile REST data model.

The model field choices are listed in the preferred display order,
most common to least common.

The data capture client has the following responsibility:

* Validate the data upon input as determined by the model
  validation documentation.

* Resolve conflicts between data capture and the model, e.g. the
  default value or validation.
"""



"""
Choice utility functions.
"""

def max_length(choices):
    """
    Returns the size of the longest choice.
    
    :param: the available choice strings or tuples
    :return: the maximum length
    """
    def length(item):
        if isinstance(item, str):
            value = item
        else:
            value = item[0]
        return len(value)
    
    return max((length(c) for c in choices))


def roman_range_choices(start, stop):
    """
    Returns the (arabic number, roman numeral) choice tuples
    for the given exclusive range bounds. This is useful, e.g.,
    for displaying the tumor stage. (Yes, this is the 21st
    century).
     
    Example:

    >>> from qirest_client import choices
    >>> choices.roman_range_choices(1, 5)
    [('I', 1), ('II', 2), ('III', 3), ('IV', 4)]
    
    :param start: the first value in the range
    :param stop: one greater than the last value in the range
    :param roman: flag indicating whether the display value
        is a roman numeral
    :return: the [(value, label)] choices list
    :raise ValueError: if the *roman* flag is set and start
        is less than one or stop is greater than five
    """
    return [(_roman(v), v) for v in range(start, stop)]


def _roman(n):
    """
    :param n: the input integer
    :return the roman numeral
    :raise ValueError: if the input integer is not a positive
        integer less than five
    """
    if n not in range(1, 5):
        raise ValueError("The roman numeral converter is not supported"
                         " for the value %d" % n)
    if n == 4:
        return 'IV'
    else:
        return 'I' * n



from qirest_client import choices
from nose.tools import assert_equal

class TestChoices(object):
    def test_roman_range_choices(self):
        expected = [('I', 1), ('II', 2), ('III', 3), ('IV', 4)]
        actual = choices.roman_range_choices(1, 5)
        assert_equal(actual, expected)


if __name__ == "__main__":
    import nose
    nose.main(defaultTest=__name__)



"""
The data model classes common to both imaging and clinical observations.
"""

import mongoengine
from mongoengine import (fields, ValidationError)


class Encounter(mongoengine.EmbeddedDocument):
    """The patient clinical encounter, e.g. biopsy or imaging session."""

    meta = dict(allow_inheritance=True)

    date = fields.DateTimeField(required=True)
    """The encounter date."""

    weight = fields.IntField()
    """The patient weight in kilograms."""


class Outcome(mongoengine.EmbeddedDocument):
    """The result of a patient clinical or imaging observation."""

    meta = dict(allow_inheritance=True)


class TumorExtent(Outcome):
    """The tumor 3D measurements in millimeters."""

    length = fields.IntField()
    """The longest tumor convex hull diameter length."""

    width = fields.IntField()
    """
    The longest tumor convex hull diameter perpendicular to the
    length segment.
    """

    depth = fields.IntField()
    """
    The longest tumor convex hull diameter perpendicular to the
    length and width segments.
    """
